import logging
from uuid import UUID

from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db import transaction
from django.db.models import Count, F, Max, Q
from django.http.response import Http404
from django.shortcuts import get_object_or_404
from rest_framework import status
from rest_framework.exceptions import ValidationError
from rest_framework.generics import (
    CreateAPIView,
    GenericAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveAPIView,
    RetrieveDestroyAPIView,
    RetrieveUpdateAPIView,
    RetrieveUpdateDestroyAPIView,
)
from rest_framework.response import Response
from rest_framework.views import APIView

from arkindex.dataimport.models import (
    DataFile,
    DataImport,
    Repository,
    RepositoryType,
    Revision,
    Worker,
    WorkerRun,
    WorkerVersion,
)
from arkindex.dataimport.serializers.files import DataFileCreateSerializer, DataFileSerializer
from arkindex.dataimport.serializers.git import ExternalRepositorySerializer, RevisionSerializer
from arkindex.dataimport.serializers.imports import (
    DataImportFromFilesSerializer,
    DataImportLightSerializer,
    DataImportSerializer,
    ElementLightSerializer,
    ElementsWorkflowSerializer,
    ImportTranskribusSerializer,
    StartProcessSerializer,
    WorkerRunLightSerializer,
    WorkerRunSerializer,
)
from arkindex.dataimport.serializers.workers import RepositorySerializer, WorkerSerializer, WorkerVersionSerializer
from arkindex.documents.api.elements import ElementsListMixin
from arkindex.documents.models import ClassificationState, Corpus, Element, ElementType
from arkindex.project.fields import ArrayRemove
from arkindex.project.mixins import CorpusACLMixin, CustomPaginationViewMixin, DeprecatedMixin, SelectionMixin
from arkindex.project.openapi import AutoSchema
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.users.models import OAuthCredentials, Role, User
from arkindex_common.enums import DataImportMode
from ponos.models import STATES_ORDERING, State

logger = logging.getLogger(__name__)


class DataImportsList(CorpusACLMixin, ListAPIView):
    """
    List all data imports
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportLightSerializer
    openapi_overrides = {
        'operationId': 'ListDataImports',
        'tags': ['imports'],
        'parameters': [
            {
                'name': 'corpus',
                'in': 'query',
                'description': 'Filter imports by corpus ID',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
            {
                'name': 'mode',
                'in': 'query',
                'description': 'Filter imports by mode',
                'required': False,
                'schema': {
                    'enum': [mode.value for mode in DataImportMode],
                },
            },
            {
                'name': 'state',
                'in': 'query',
                'description': 'Filter imports by workflow state',
                'required': False,
                'schema': {
                    'enum': [state.value for state in State],
                },
            },
            {
                'name': 'id',
                'in': 'query',
                'description': 'Filter imports by beginning of UUID',
                'required': False,
                'schema': {
                    'type': 'string',
                },
            },
            {
                'name': 'with_workflow',
                'in': 'query',
                'description': 'Restrict to or exclude import with workflow',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': True
                },
            }
        ]
    }

    def get_queryset(self):
        with_workflow = self.request.query_params.get('with_workflow')
        filters = Q(
            Q(
                Q(corpus__isnull=True)
                | Q(corpus__in=Corpus.objects.readable(self.request.user))
            )
            & Q(workflow__isnull=bool(with_workflow and with_workflow.lower() in ('false', '0')))
        )

        if 'corpus' in self.request.query_params:
            corpus_id = self.request.query_params['corpus']
            try:
                corpus_id = UUID(corpus_id)
            except (AttributeError, ValueError):
                raise ValidationError({'corpus': "'{}' is not a valid UUID".format(corpus_id)})
            filters &= Q(corpus=self.get_corpus(corpus_id))

        if 'mode' in self.request.query_params:
            try:
                filters &= Q(mode=DataImportMode(self.request.query_params['mode']))
            except ValueError:
                raise ValidationError({
                    'mode': "Mode '{}' does not exist".format(self.request.query_params['mode']),
                })

        if 'creator' in self.request.query_params:
            try:
                user_filters = {'id': int(self.request.query_params['creator'])}
            except ValueError:
                user_filters = {'email': self.request.query_params['creator']}

            try:
                filters &= Q(creator=User.objects.get(**user_filters))
            except User.DoesNotExist:
                raise ValidationError({
                    'creator': "User with email or id '{}' does not exist".format(self.request.query_params['creator']),
                })

        if 'id' in self.request.query_params:
            filters &= Q(id__startswith=self.request.query_params['id'])

        qs = DataImport.objects.filter(filters).prefetch_related('workflow__tasks')
        # Order workflow by date of last updated task in workflow
        qs = qs.annotate(
            last_task=Max('workflow__tasks__updated'),
            last_run=Max('workflow__tasks__run')
        )

        state_value = self.request.query_params.get('state')
        if state_value :
            try:
                state = State(self.request.query_params['state'])
            except ValueError:
                raise ValidationError({'state': f"State '{state_value}' does not exist"})
            # Filter out processes which have a task with an incompatible state on their last run
            excluding_states = STATES_ORDERING[:STATES_ORDERING.index(state)]
            excluded_imports = qs.filter(
                Q(workflow__tasks__run=F('last_run')),
                Q(workflow__tasks__state__in=excluding_states)
            )
            # Keep non excluded processes matching the state on their last run tasks
            state_query = Q(workflow__tasks__run=F('last_run')) & Q(workflow__tasks__state=state)
            if state == State.Unscheduled:
                # Handle a workflow with no task as unscheduled
                state_query |= Q(workflow__tasks__isnull=True)
            qs = qs.filter(state_query).exclude(id__in=excluded_imports.values('id')).distinct()

        return qs.order_by('-last_task')


class DataImportDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve or edit a process
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return DataImport.objects.filter(
            Q(corpus__isnull=True)
            | Q(corpus__in=Corpus.objects.readable(self.request.user))
        ).select_related('corpus', 'workflow').prefetch_related('workflow__tasks').annotate(last_run=Max('workflow__tasks__run'))

    def get_object(self):
        if not hasattr(self, 'dataimport'):
            self.element = super().get_object()
        return self.element

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['dataimport'] = self.get_object()
        return context

    def check_object_permissions(self, request, instance):
        super().check_object_permissions(request, instance)
        if request.method == 'GET':
            return

        if not instance.corpus:
            user = request.user
            # Fall back basic user checks
            if user != instance.creator and not user.is_admin:
                self.permission_denied(request, message='You are not the creator of this process nor an admin.')
        elif not self.has_write_access(instance.corpus):
            self.permission_denied(request, message=f'You have no write access to corpus "{instance.corpus.name}".')
        if request.method == 'DELETE' and instance.state == State.Running:
            raise ValidationError({'__all__': ['Cannot delete a workflow while it is running']})


class DataImportRetry(CorpusACLMixin, GenericAPIView):
    """
    Retry a data import. Can only be used on imports with Error, Failed or Stopped states.
    """
    permission_classes = (IsVerified, )
    serializer_class = DataImportSerializer
    openapi_overrides = {
        'operationId': 'RetryDataImport',
        'tags': ['imports'],
    }

    def get_queryset(self):
        return DataImport.objects.filter(
            Q(corpus__isnull=True)
            | Q(corpus__in=Corpus.objects.writable(self.request.user))
        )

    def post(self, request, *args, **kwargs):
        instance = self.get_object()
        if not instance.corpus:
            user = self.request.user
            # Fall back basic user checks
            if user is not instance.creator and not user.is_admin:
                raise PermissionDenied
        if instance.workflow is not None and instance.state in (State.Unscheduled, State.Pending):
            raise ValidationError('This workflow is already pending')
        elif instance.state == State.Running:
            raise ValidationError('This workflow is already running')
        instance.retry()
        serializer = self.get_serializer(instance)
        return Response(serializer.data)


class DataImportFromFiles(CreateAPIView):
    """
    Start a data import from one or more uploaded files.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataImportFromFilesSerializer
    openapi_overrides = {
        'tags': ['files'],
    }

    def create(self, *args, **kwargs):
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        files = serializer.validated_data['files']
        corpus = files[0].corpus
        mode = serializer.validated_data['mode']
        # Serializer validation codes returns an element for 'folder_id'
        folder = serializer.validated_data.get('folder_id')
        folder_type = serializer.validated_data.get('folder_type')
        element_type = serializer.validated_data['element_type']

        if folder and folder.corpus != corpus:
            # The files' corpus is already validated as writable
            raise ValidationError('Element and files are in different corpora')

        # Start a dataimport with thumbnails generation
        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=mode,
            element=folder,
            folder_type=folder_type,
            element_type=element_type,
        )
        self.dataimport.files.set(files)
        self.dataimport.start(thumbnails=True)


class CorpusWorkflow(SelectionMixin, CreateAPIView):
    """
    Start a Worker Process from Arkindex corpus elements
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementsWorkflowSerializer
    openapi_overrides = {
        'operationId': 'CreateElementsWorkflow',
        'description': 'Create a distributed workflow from elements of an Arkindex corpus',
        'tags': ['imports'],
        'responses': {
            '201': {
                'content': {
                    'application/json': {
                        'schema': AutoSchema()._map_serializer(DataImportSerializer())
                    }
                }
            }
        }
    }

    def create(self, request, pk=None, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        corpus, element, process_name, name, elt_type, selection, best_class, load_children = map(
            lambda key: serializer.validated_data.pop(key, None),
            (
                'corpus',
                'element',
                'process_name',
                'name',
                'type',
                'selection',
                'best_class',
                'load_children'
            )
        )

        if selection:
            elements = self.get_selection(corpus.id)
            if not elements:
                raise ValidationError({
                    '__all__': ['No element match filtering']
                })

        element_type = None
        if elt_type:
            element_type = ElementType.objects.get(corpus=corpus, slug=elt_type)

        dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Workers,
            name=process_name,
            name_contains=name,
            element_type=element_type,
            element=element,
            best_class=best_class,
            load_children=load_children
        )

        if selection:
            dataimport.elements.set(elements)

        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(
                dataimport,
                context=self.get_serializer_context()
            ).data,
        )


class StartProcess(APIView):
    """
    Start a process, used to build a Workflow with Workers
    """
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'StartProcess',
        'tags': ['imports'],
    }

    def post(self, request, pk=None, **kwargs):
        dataimport = get_object_or_404(DataImport, pk=self.kwargs['pk'])

        if dataimport.mode != DataImportMode.Workers or dataimport.workflow is not None:
            raise ValidationError('Only a DataImport with Workers mode and not already launched can be started later on')

        serializer = StartProcessSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        dataimport.start(**serializer.validated_data)

        return Response(
            status=status.HTTP_200_OK,
            data=DataImportSerializer(
                dataimport,
                context={'request': request}
            ).data,
        )


class DataImportElements(DeprecatedMixin, ListAPIView):
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListDataImportElements',
        'tags': ['imports'],
        'description': "List a DataImport's associated elements, "
                       "for DataImports running workers on existing elements.",
    })
    deprecation_message = 'Listing elements on a DataImport is now implemented through ListProcessElements.'


class DataFileList(CorpusACLMixin, ListAPIView):
    """
    List uploaded files in a corpus
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    # Tell the OpenAPI schema generator to believe this view is a list
    action = 'List'
    openapi_overrides = {
        'tags': ['files']
    }

    def get_queryset(self):
        return DataFile.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class DataFileRetrieve(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get one file
    """
    permission_classes = (IsVerified, )
    serializer_class = DataFileSerializer
    openapi_overrides = {
        'tags': ['files'],
    }

    def get_queryset(self):
        # We need to use the default database to avoid stale read
        # when a datafile is updated immediately after it's created on a small upload
        return DataFile.objects.using('default').filter(
            corpus__in=Corpus.objects.readable(self.request.user)
        )

    def perform_update(self, serializer):
        if not self.has_write_access(serializer.instance.corpus):
            raise PermissionDenied
        return super().perform_update(serializer)

    def perform_destroy(self, instance):
        if not self.has_write_access(instance.corpus):
            raise PermissionDenied
        try:
            instance.s3_delete()
        except Exception as e:
            logger.warning(f'Could not delete DataFile from S3: {e}')
        return super().perform_destroy(instance)


class DataFileUpload(DeprecatedMixin, CreateAPIView):
    """
    Upload a new file to a corpus
    """
    openapi_overrides = {
        'operationId': 'UploadDataFile',
        'tags': ['files'],
    }
    deprecation_message = 'DataFile uploads via the backend are now deprecated. ' \
        'Please use `CreateDataFile`, upload to S3 using the returned `s3_put_url`, ' \
        'then use `PartialUpdateDataFile` to set its status to `checked`.'


class DataFileCreate(CreateAPIView):
    """
    Create a Datafile.
    In case of success, a signed URI is returned to upload file content directly to a remote server.
    """

    permission_classes = (IsVerified, )
    serializer_class = DataFileCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateDataFile',
        'tags': ['files'],
    }


class GitRepositoryImportHook(APIView):
    """
    This endpoint is intended as a webhook for Git repository hosting applications like GitLab.
    """
    openapi_overrides = {
        'operationId': 'GitPushHook',
        'security': [],
        'tags': ['repos'],
    }

    def post(self, request, pk=None, **kwargs):
        repo = get_object_or_404(Repository, id=pk)
        assert repo.enabled, 'No credentials available'
        repo.provider_class(credentials=repo.credentials).handle_webhook(repo, request)
        return Response(status=status.HTTP_204_NO_CONTENT)


class RepositoryList(ListAPIView):
    """
    List connected repositories
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer
    openapi_overrides = {
        'operationId': 'ListRepositories',
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Repository.objects \
            .filter(
                Q(corpora__isnull=True)
                | Q(corpora__in=Corpus.objects.readable(self.request.user))
            ) \
            .distinct() \
            .select_related('credentials') \
            .prefetch_related('corpora', 'workers') \
            .order_by('url')


class AvailableRepositoriesList(ListCreateAPIView):
    """
    List repositories associated to user OAuth credentials
    Using the given OAuth credentials ID, this uses the Git hosting
    application API's search feature to look for a repository matching
    the given query. Without a query, returns a full list.

    A POST request allow to create the repository in Arkindex and start a dataimport
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = ExternalRepositorySerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        cred = get_object_or_404(OAuthCredentials, user=self.request.user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repos = provider.list_repos(query=self.request.query_params.get('search'))
        return map(provider.to_dict, repos)

    @transaction.atomic
    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        user = self.request.user
        cred = get_object_or_404(OAuthCredentials, user=user, id=self.kwargs['pk'])
        provider = cred.git_provider_class(credentials=cred)
        repo = provider.create_repo(
            request=self.request,
            id=serializer.validated_data.get('id')
        )
        latest_commit_sha = provider.get_latest_commit_sha(repo)
        rev, _ = provider.get_or_create_revision(repo, latest_commit_sha)
        extra_kwargs = {}

        # In case of a IIIF import, a corpus should be created and referenced to this repository
        if repo.type == RepositoryType.IIIF:
            corpus = Corpus.objects.create(
                name=f'IIIF import from {repo.url}',
                description=f'Automatic import from IIIF repository ({repo.url})',
                repository=repo
            )
            # Give the creator a write right on the corpus
            corpus.memberships.create(user=user, level=Role.Admin.value)
            # Add default types
            corpus.create_default_types()
            extra_kwargs['corpus_id'] = corpus.id

        dataimport = DataImport.objects.create(
            revision=rev,
            mode=DataImportMode.Repository,
            creator=cred.user,
            **extra_kwargs
        )
        # Generate thumbnails for IIIF repositories revisions only
        gen_thumbnails = repo.type == RepositoryType.IIIF
        dataimport.start(thumbnails=gen_thumbnails, **extra_kwargs)

        # Return the serialized data import
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(dataimport, context={'request': self.request}).data,
        )


class RepositoryRetrieve(CorpusACLMixin, RetrieveDestroyAPIView):
    """
    Retrieve a repository
    """
    permission_classes = (IsVerified, )
    serializer_class = RepositorySerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Repository.objects.filter(
            Q(corpora__isnull=True)
            | Q(corpora__in=Corpus.objects.readable(self.request.user))
        ).distinct()

    def perform_destroy(self, instance):
        user = self.request.user
        if user != instance.credentials.user and not user.is_admin:
            raise PermissionDenied
        if not all(self.has_write_access(c) for c in instance.corpora.all()):
            raise PermissionDenied
        return super().perform_destroy(instance)


class RevisionRetrieve(CorpusACLMixin, RetrieveAPIView):
    """
    Retrieve a revision
    """
    permission_classes = (IsVerified, )
    serializer_class = RevisionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Revision.objects.filter(
            Q(repo__corpora__isnull=True)
            | Q(repo__corpora__in=Corpus.objects.readable(self.request.user))
        )


class MLToolList(DeprecatedMixin, ListAPIView):
    """
    List available machine learning tools
    """
    openapi_overrides = {
        'operationId': 'ListMLTools',
        'security': [],
        'tags': ['ml'],
    }
    deprecation_message = 'Listing ML Tools is now deprecated, please use the new worker system.'


class WorkerList(ListCreateAPIView):
    """
    List workers for a given repository UUID

    Create a Worker instance on a repository for given slug, name and type, return existing
    query if a worker with the same slug already exists on the repository
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = WorkerSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Worker.objects.filter(repository_id=self.kwargs['pk'])

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        repo = get_object_or_404(Repository, pk=self.kwargs['pk'])

        worker, created = repo.workers.get_or_create(
            slug=serializer.validated_data['slug'],
            defaults={
                'name': serializer.validated_data['name'],
                'type': serializer.validated_data['type'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerSerializer(worker).data, status=reponse_status)


class WorkerVersionList(ListCreateAPIView):
    """
    List versions, their revision and associated git references for a given worker UUID

    Create a WorkerVersion instance for a worker and a revision with a given JSON configuration,
    return existing query if a workerVersion already exists for this worker and this revision
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = WorkerVersionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return WorkerVersion.objects \
            .filter(worker_id=self.kwargs['pk']) \
            .select_related('revision__repo', 'worker__repository') \
            .prefetch_related('revision__refs', 'revision__versions') \
            .order_by('-revision__created')

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        # We need to use the default database to avoid stale read
        # when creating a worker version on a newly created worker
        try:
            worker = Worker.objects.using('default').get(pk=self.kwargs['pk'])
        except Worker.DoesNotExist:
            raise Http404

        version, created = WorkerVersion.objects.get_or_create(
            worker=worker,
            revision=serializer.validated_data['revision'],
            defaults={
                'configuration': serializer.validated_data['configuration'],
            }
        )

        reponse_status = status.HTTP_201_CREATED if created else status.HTTP_200_OK

        return Response(WorkerVersionSerializer(version).data, status=reponse_status)


class CorpusWorkerVersionList(CorpusACLMixin, ListAPIView):
    """
    List worker versions used by elements of a given corpus.
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    pagination_class = None
    serializer_class = WorkerVersionSerializer
    openapi_overrides = {
        'operationId': 'ListCorpusWorkerVersions',
        'tags': ['ml'],
    }

    def get_queryset(self):
        return WorkerVersion.objects \
            .filter(elements__corpus=self.get_corpus(self.kwargs['pk'])) \
            .select_related('revision__repo', 'worker__repository') \
            .prefetch_related('revision__refs', 'revision__versions') \
            .order_by('-revision__created') \
            .annotate(element_count=Count('id'))


class WorkerVersionRetrieve(CorpusACLMixin, RetrieveUpdateAPIView):
    """
    Retrieve a specific worker version
    """
    permission_classes = (IsVerifiedOrReadOnly, )
    serializer_class = WorkerVersionSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return WorkerVersion.objects.filter(
            Q(revision__repo__corpora__isnull=True)
            | Q(revision__repo__corpora__in=Corpus.objects.readable(self.request.user))
        )


class WorkerRetrieve(RetrieveAPIView):
    """
    Retrieve a specific worker
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerSerializer
    openapi_overrides = {
        'tags': ['repos'],
    }

    def get_queryset(self):
        return Worker.objects.filter(
            Q(repository__corpora__isnull=True)
            | Q(repository__corpora__in=Corpus.objects.readable(self.request.user))
        )


class WorkerRunList(ListCreateAPIView):
    """
    List worker runs for a given dataimport UUID
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunLightSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return WorkerRun.objects.filter(
            dataimport_id=self.kwargs['pk']
        ).order_by('id')

    def perform_create(self, serializer):
        dataimport = get_object_or_404(DataImport, pk=self.kwargs['pk'])

        if dataimport.mode != DataImportMode.Workers:
            raise ValidationError({'dataimport': ['Import mode must be Workers']})

        if dataimport.workflow is not None:
            raise ValidationError({'dataimport': ["Can't create a WorkerRun on a DataImport that has already started"]})

        serializer.save(dataimport=dataimport)


class WorkerRunDetails(CorpusACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve a specific worker run
    """
    permission_classes = (IsVerified, )
    serializer_class = WorkerRunSerializer
    openapi_overrides = {
        'tags': ['imports'],
    }

    def get_queryset(self):
        return WorkerRun.objects.filter(
            Q(dataimport__corpus__isnull=True)
            | Q(dataimport__corpus__in=Corpus.objects.readable(self.request.user))
        )

    def perform_destroy(self, instance):
        # Remove references to this instance from its children parents field
        instance.dataimport.worker_runs.filter(parents__contains=[instance.id]).update(parents=ArrayRemove('parents', instance.id))

        return super().perform_destroy(instance)


class ImportTranskribus(CreateAPIView):
    """
    Start a data import from Transkribus email and collection ID.
    """

    permission_classes = (IsVerified, )
    serializer_class = ImportTranskribusSerializer
    openapi_overrides = {
        'operationId': 'CreateImportTranskribus',
        'description': 'Create a data import from Transkribus collection ID.',
        'tags': ['imports'],
        'responses': {
            '201': {
                'content': {
                    'application/json': {
                        'schema': AutoSchema()._map_serializer(DataImportSerializer())
                    }
                }
            }
        }
    }

    def create(self, *args, **kwargs):
        if not settings.ARKINDEX_FEATURES['transkribus']:
            raise ValidationError(['Transkribus import is unavailable due to the transkribus feature being disabled.'])
        super().create(*args, **kwargs)
        return Response(
            status=status.HTTP_201_CREATED,
            data=DataImportSerializer(self.dataimport, context={'request': self.request}).data,
        )

    def perform_create(self, serializer):
        collection_id = serializer.validated_data['collection_id']

        # Create corpus
        corpus = Corpus.objects.create(
            name=f"Transkribus collection n°{collection_id}",
        )
        corpus.memberships.create(user=self.request.user, level=Role.Admin.value)
        corpus.create_default_transkribus_types()

        # Start a dataimport with thumbnails generation
        self.dataimport = corpus.imports.create(
            creator=self.request.user,
            mode=DataImportMode.Transkribus,
            collection_id=collection_id,
            build_entities=serializer.validated_data['build_entities']
        )
        self.dataimport.start(thumbnails=True)


class ListProcessElements(CustomPaginationViewMixin, ListAPIView):
    """
    List all elements for a specific process
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementLightSerializer
    openapi_overrides = {
        'operationId': 'ListProcessElements',
        'tags': ['imports']
    }

    def retrieve_elements(self, dataimport):
        elements = None

        if dataimport.element:
            # Handle a single element
            if dataimport.element.corpus_id != dataimport.corpus_id:
                raise ValidationError({
                    'element_id': [
                        'Element {} is not part of corpus {}'.format(
                            dataimport.element.id, dataimport.corpus.id
                        )
                    ]
                })
            elements = Element.objects.filter(id=dataimport.element.id)

        elif dataimport.elements.exists():
            # Handle a selection of elements
            # The list() is necessary to make a unique SQL request to check length & content
            # It's more performant, as we usually have only 1 corpus in here
            elements = dataimport.elements.all()
            corpus = list(elements.values_list('corpus_id', flat=True).distinct())
            # Check all elements are in the same corpus as the process
            if len(corpus) != 1 or corpus[0] != dataimport.corpus_id:
                raise ValidationError({
                    'elements': [
                        'Some elements on this process are not part of corpus {}'.format(
                            dataimport.corpus.id
                        )]
                })

        # Apply base filters as early as possible to trim results
        base_filters = self.get_filters(dataimport)

        if elements is not None and dataimport.load_children:
            # Load all the children elements whose path contains the pre-selected elements
            # Those children are appended to the pre-selection
            elements |= Element.objects.filter(
                paths__path__overlap=map(str, elements.values_list('id', flat=True)),
                **base_filters
            )

        # Load the full corpus, only when elements has not been populated before
        if elements is None:
            # Handle all elements of the process corpus
            elements = Element.objects.filter(corpus=dataimport.corpus_id)

        # Filter elements depending on process properties
        elements = elements.filter(**base_filters)
        class_filters = self.get_classifications_filters(dataimport)
        if class_filters is not None:
            elements = elements.filter(class_filters).distinct()

        # Only retrieve necessary values for the serializer
        return elements.values('id', 'type__slug', 'name')

    def get_filters(self, dataimport):
        filters = {
            "corpus_id": dataimport.corpus_id,
        }
        if dataimport.name_contains:
            filters['name__contains'] = dataimport.name_contains
        if dataimport.element_type:
            filters['type_id'] = dataimport.element_type_id
        else:
            # Limit the scope of types available to merge
            # This prevent memory from exploding when no type is selected
            filters['type__corpus_id'] = dataimport.corpus_id

        return filters

    def get_classifications_filters(self, dataimport):
        if dataimport.best_class is None:
            return

        # Generic ORM query to find best classes:
        # - elements with a validated classification
        # - OR where high confidence is True
        best_classifications = Q(classifications__state=ClassificationState.Validated) \
            | Q(classifications__high_confidence=True)

        # List elements without any best classes, by inverting the query above
        if dataimport.best_class in ('false', '0'):
            return ~best_classifications

        try:
            # Filter on a specific class
            class_filter = UUID(dataimport.best_class)
            return best_classifications & Q(classifications__ml_class_id=class_filter)
        except (TypeError, ValueError):
            # By default, use all best classifications
            return best_classifications

    def get_queryset(self):
        dataimport = get_object_or_404(
            # Avoid stale read on newly created dataimports
            DataImport.objects.all().using('default'),
            Q(pk=self.kwargs['pk'])
            & Q(
                Q(corpus__isnull=False)
                & Q(corpus__in=Corpus.objects.readable(self.request.user))
            )
        )

        if dataimport.mode not in (DataImportMode.Elements, DataImportMode.Workers):
            return Element.objects.none()

        return self.retrieve_elements(dataimport)
