from unittest.mock import patch

import yaml
from django.conf import settings
from django.test import override_settings
from django.urls import reverse
from rest_framework import status

from arkindex.dataimport.models import DataImport
from arkindex.project.default_corpus import DEFAULT_TRANSKRIBUS_TYPES
from arkindex.project.tests import FixtureAPITestCase
from arkindex.users.models import Role, User
from arkindex_common.enums import DataImportMode
from ponos.models import State


@override_settings(TRANSKRIBUS_EMAIL="arkindex@teklia.com", TRANSKRIBUS_PASSWORD=None, TRANSKRIBUS_WORKER_VERSION="12345-12345")
class TestTranskribusImport(FixtureAPITestCase):
    """
    Test transkribus import
    """

    @classmethod
    def setUpTestData(cls):
        super().setUpTestData()
        User.objects.update(transkribus_email="nope@nope.fr")

    def test_requires_login(self):
        response = self.client.post(reverse("api:import-transkribus"))
        self.assertEqual(response.status_code, status.HTTP_403_FORBIDDEN)

    @override_settings(ARKINDEX_FEATURES={"transkribus": False})
    def test_requires_flag(self):
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), ["Transkribus import is unavailable due to the transkribus feature being disabled."])

    def test_requires_transkribus_email(self):
        User.objects.update(transkribus_email=None)
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"__all__": ["You have not register your transkribus email"]})

    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_arkindex_has_not_access(self, mock_transkribus):
        # Not a mistake: Transkribus client raises `Exception` directly when it runs out of retries
        mock_transkribus.side_effect = Exception("401 Unauthorized")
        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"collection_id" : ["User arkindex@teklia.com is not a member of the collection 12345"]})

    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_user_has_not_access(self, mock_transkribus):
        mock_transkribus.return_value = [{"email": "arkindex@teklia.com"}]

        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_400_BAD_REQUEST)
        self.assertEqual(response.json(), {"collection_id": ["User nope@nope.fr is not a member of the collection 12345"]})

    @override_settings(
        PONOS_RECIPE={}
    )
    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_create_import(self, mock_transkribus):
        mock_transkribus.return_value = [{"email": "nope@nope.fr"}, {"email": "arkindex@teklia.com"}]

        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data["id"])
        self.assertEqual(dataimport.mode, DataImportMode.Transkribus)
        self.assertEqual(dataimport.collection_id, 12345)
        corpus = dataimport.corpus
        right = corpus.memberships.get(user=self.user)
        self.assertTrue(right.level >= Role.Admin.value)
        self.assertEqual(corpus.name, "Transkribus collection n°12345")
        self.assertEqual(corpus.description, "")
        self.assertEqual(corpus.public, False)
        # Assert defaults types are set on the new corpus
        self.assertCountEqual(
            list(corpus.types.values(
                "slug",
                "display_name",
                "folder",
            )),
            [{
                "folder": False,
                **values
            } for values in DEFAULT_TRANSKRIBUS_TYPES]
        )
        self.assertEqual(dataimport.state, State.Unscheduled)
        self.assertIsNotNone(dataimport.workflow)
        recipe = yaml.safe_load(dataimport.workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_PROCESS_ID': str(dataimport.id),
                'TRANSKRIBUS_EMAIL': settings.TRANSKRIBUS_EMAIL,
                'TRANSKRIBUS_PASSWORD': settings.TRANSKRIBUS_PASSWORD,
                'TRANSKRIBUS_WORKER_VERSION': settings.TRANSKRIBUS_WORKER_VERSION
            },
            'tasks': {
                'import': {
                    'command': 'python -m arkindex_tasks.import_transkribus {} --corpus {}'.format(dataimport.collection_id, corpus.id),
                    'image': 'registry.gitlab.com/arkindex/tasks'
                },
                'thumbnails': {
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import/elements.json',
                    'image': 'registry.gitlab.com/arkindex/tasks',
                    'parents': ['import']
                }
            }
        })

    @override_settings(
        PONOS_RECIPE={}
    )
    @patch("transkribus.TranskribusAPI.list_user_collection")
    def test_create_import_build_entities(self, mock_transkribus):
        mock_transkribus.return_value = [{"email": "nope@nope.fr"}, {"email": "arkindex@teklia.com"}]

        self.client.force_login(self.user)
        response = self.client.post(reverse("api:import-transkribus"), {
            "collection_id": "12345",
            "build_entities": True
        }, format="json")
        self.assertEqual(response.status_code, status.HTTP_201_CREATED)
        data = response.json()
        dataimport = DataImport.objects.get(id=data["id"])
        self.assertEqual(dataimport.mode, DataImportMode.Transkribus)
        self.assertEqual(dataimport.collection_id, 12345)
        corpus = dataimport.corpus
        right = corpus.memberships.get(user=self.user)
        self.assertTrue(right.level >= Role.Admin.value)
        self.assertEqual(corpus.name, "Transkribus collection n°12345")
        self.assertEqual(corpus.description, "")
        self.assertEqual(corpus.public, False)
        # Assert defaults types are set on the new corpus
        self.assertCountEqual(
            list(corpus.types.values(
                "slug",
                "display_name",
                "folder",
            )),
            [{
                "folder": False,
                **values
            } for values in DEFAULT_TRANSKRIBUS_TYPES]
        )
        self.assertEqual(dataimport.state, State.Unscheduled)
        self.assertIsNotNone(dataimport.workflow)
        recipe = yaml.safe_load(dataimport.workflow.recipe)
        self.assertDictEqual(recipe, {
            'env': {
                'ARKINDEX_PROCESS_ID': str(dataimport.id),
                'TRANSKRIBUS_EMAIL': settings.TRANSKRIBUS_EMAIL,
                'TRANSKRIBUS_PASSWORD': settings.TRANSKRIBUS_PASSWORD,
                'TRANSKRIBUS_WORKER_VERSION': settings.TRANSKRIBUS_WORKER_VERSION
            },
            'tasks': {
                'import': {
                    'command': 'python -m arkindex_tasks.import_transkribus {} --corpus {}'.format(dataimport.collection_id, corpus.id),
                    'image': 'registry.gitlab.com/arkindex/tasks'
                },
                'build_entities': {
                    'command': 'python -m arkindex_tasks.build_entities /data/import/{}/transcriptions.json'.format(dataimport.collection_id),
                    'image': 'registry.gitlab.com/arkindex/tasks',
                    'parents': ['import']
                },
                'thumbnails': {
                    'command': 'python3 -m arkindex_tasks.generate_thumbnails /data/import/elements.json',
                    'image': 'registry.gitlab.com/arkindex/tasks',
                    'parents': ['import']
                }
            }
        })
