import email.utils
import uuid
from collections import defaultdict
from datetime import datetime, timezone

from django.conf import settings
from django.db import connection, transaction
from django.db.models import CharField, Count, Max, Prefetch, Q, QuerySet
from django.db.models.functions import Cast
from django.shortcuts import get_object_or_404
from django.utils.functional import cached_property
from psycopg2.extras import execute_values
from rest_framework import permissions, serializers, status
from rest_framework.exceptions import NotFound, PermissionDenied, ValidationError
from rest_framework.generics import (
    CreateAPIView,
    DestroyAPIView,
    ListAPIView,
    ListCreateAPIView,
    RetrieveUpdateDestroyAPIView,
    UpdateAPIView,
)
from rest_framework.mixins import DestroyModelMixin
from rest_framework.response import Response

from arkindex.dataimport.models import WorkerVersion
from arkindex.documents.models import (
    AllowedMetaData,
    Classification,
    ClassificationState,
    Corpus,
    Element,
    ElementPath,
    ElementType,
    MetaData,
    Transcription,
)
from arkindex.documents.serializers.elements import (
    CorpusSerializer,
    ElementBulkSerializer,
    ElementCreateSerializer,
    ElementListSerializer,
    ElementNeighborsSerializer,
    ElementParentSerializer,
    ElementSerializer,
    ElementSlimSerializer,
    ElementTypeSerializer,
    MetaDataUpdateSerializer,
)
from arkindex.documents.serializers.light import CorpusAllowedMetaDataSerializer, ElementTypeLightSerializer
from arkindex.documents.serializers.ml import ElementTranscriptionSerializer
from arkindex.images.models import Zone
from arkindex.project.mixins import ACLMixin, CorpusACLMixin, SelectionMixin
from arkindex.project.openapi import AutoSchema
from arkindex.project.pagination import PageNumberPagination
from arkindex.project.permissions import IsVerified, IsVerifiedOrReadOnly
from arkindex.project.tools import BulkMap
from arkindex.project.triggers import corpus_delete, element_trash
from arkindex.users.models import Role
from arkindex.users.utils import filter_rights

classifications_queryset = Classification.objects.select_related('ml_class', 'worker_version').order_by('-confidence')

best_classifications_prefetch = Prefetch(
    'classifications',
    queryset=classifications_queryset
    # Keep best or validated distinct classifications
    .filter(Q(state=ClassificationState.Validated) | Q(high_confidence=True))
    # Remove rejected classifications
    .exclude(state=ClassificationState.Rejected).distinct(),
    to_attr='best_classes'
)


def _fetch_has_children(elements):
    """
    From a list of elements, sets the `has_children` attribute to a boolean
    indicating if an element has children elements.

    The initial goal was to have a way for the frontend to tell if an element had any children;
    having a count was optional. It is possible to just use an Exists subquery since Django 2.2:
    Element.objects.annotate(has_children=Exists(ElementPath.objects.filter(path__last=OuterRef('pk'))))
    This works but makes the element list queries 20 times slower, due to the way the Postgres query planner
    handles this.

    We used a second query with a GROUP BY to retrieve the element path count, because the difference between
    telling if paths exists and getting a count is negligible with a GROUP BY, but the query could get very slow
    for elements with a lot of children, even without the count.

    We switched back to a boolean and to a query using a LEFT JOIN on a set of values, using the array overlap
    operator (&&) to match any child (not just direct children). This assumes that ElementPath is properly maintained
    and that no "ghost elements" exist in the paths.
    With an array instead of values, Postgres has to resort to a very slow sequential scan over the whole table.
    With direct children (the last element in a path), Postgres cannot use the GIN index on `path` and
    uses a sequential scan and a hash join.

    The ElementsListMixin uses arkindex.project.tools.BulkMap to apply this method and perform the second request
    *after* DRF's pagination, because there is no way to perform post-processing after pagination in Django
    without having to use Django private methods.
    """
    # Ignore empty results
    if not elements:
        return elements

    with connection.cursor() as cursor:
        execute_values(
            cursor,
            'SELECT DISTINCT ON (e.id) e.id, p.id is not null as has_children '
            'FROM (VALUES %s) e (id) '
            'LEFT JOIN documents_elementpath p ON (ARRAY[e.id] && p.path)',
            tuple((element.id, ) for element in elements),
        )
        has_children = dict(cursor.fetchall())

    for element in elements:
        element.has_children = has_children[element.id]

    return elements


class ElementsListMixin(object):
    """
    Provides common features across element list endpoints (list, parents, children)
    """
    serializer_class = ElementListSerializer
    queryset = Element.objects.all()
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'type',
                'in': 'query',
                'description': 'Filter elements by type',
                'required': False,
                'schema': {
                    'type': 'string',
                },
            },
            {
                'name': 'name',
                'in': 'query',
                'description': 'Only include elements whose name contains a given string (case-insensitive)',
                'required': False,
                'schema': {'type': 'string'},
            },
            {
                'name': 'best_class',
                'in': 'query',
                'description': 'Restrict to or exclude elements with best classes, '
                               'or restrict to elements with a specific best class',
                'required': False,
                'schema': {'anyOf': [
                    {'type': 'string', 'format': 'uuid'},
                    {'type': 'boolean'},
                ]},
                'x-methods': ['get'],
            },
            {
                'name': 'folder',
                'in': 'query',
                'description': 'Restrict to or exclude elements with folder types',
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            },
            {
                'name': 'worker_version',
                'in': 'query',
                'description': 'Only include elements created by a specific worker version',
                'required': False,
                'schema': {
                    'type': 'string',
                    'format': 'uuid',
                }
            },
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False,
                },
                'x-methods': ['get'],
            },
            {
                'name': 'with_has_children',
                'in': 'query',
                'description': (
                    'Include the `has_children` boolean to tell if each element has direct children. '
                    'Otherwise, `has_children` will always be null.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                },
                'x-methods': ['get'],
            },
            {
                'name': 'with_zone',
                'in': 'query',
                'description': (
                    'Returns zone attribute for each element. '
                    'If not set, elements zone field will be returned by default.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': True
                },
                'x-methods': ['get'],
            },
            {
                'name': 'with_corpus',
                'in': 'query',
                'description': (
                    'Returns corpus attribute for each element. '
                    'If not set, elements corpus field will be returned by default.'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': True
                },
                'x-methods': ['get'],
            },
            {
                'name': 'If-Modified-Since',
                # Name for APIStar because dashes in Python kwargs would not go well
                'x-name': 'modified_since',
                'required': False,
                'in': 'header',
                'description': (
                    'Return HTTP 304 Not Modified when no elements have been created or updated since that time, '
                    'otherwise respond normally. Expects HTTP date formats (`Wed, 21 Oct 2015 07:28:00 GMT`). '
                    'Note that this is unable to detect deletions and will not filter the returned elements.'
                ),
                'schema': {
                    # Cannot use format: date-time here as HTTP headers do not use ISO 8601
                    'type': 'string'
                },
                'x-methods': ['get'],
            },
            {
                'name': 'delete_children',
                'required': False,
                'in': 'query',
                'description': 'Delete all child elements of those elements recursively.',
                'schema': {
                    'type': 'boolean',
                    'default': True
                },
                'x-methods': ['delete'],
            }
        ]
    }

    def initial(self, *args, **kwargs):
        """
        Adds a `self.clean_params` attribute that holds any query params or headers,
        filtered by their allowed HTTP method.
        Raise HTTP 400 early if any filter is not allowed in this method.
        Does not perform validation on the filter values.
        """
        super().initial(*args, **kwargs)

        self.clean_params = {}
        errors = {}
        for param in self.openapi_overrides['parameters']:
            if param['in'] == 'header':
                origin_dict = self.request.headers
            elif param['in'] == 'query':
                origin_dict = self.request.query_params
            else:
                raise NotImplementedError

            if param['name'] not in origin_dict:
                continue

            if param.get('x-methods') and self.request.method.lower() not in param['x-methods']:
                errors[param['name']] = [f'This parameter is not allowed in {self.request.method} requests.']
            else:
                self.clean_params[param['name']] = origin_dict[param['name']]

        if errors:
            raise ValidationError(errors)

    @cached_property
    def selected_corpus(self):
        corpus_id = self.kwargs.get('corpus')
        if corpus_id is None:
            return

        # When a corpus is specified, check independently that it's readable/writable
        # by the current user. This prevents comparing the corpus of every element.
        # Require write rights for 'unsafe' methods (POST/PUT/PATCH/DELETE)
        return self.get_corpus(
            corpus_id,
            role=Role.Guest if self.request.method in permissions.SAFE_METHODS else Role.Contributor
        )

    def get_filters(self, corpus=None):
        filters = {}
        if self.selected_corpus or corpus:
            filters['corpus'] = corpus or self.selected_corpus
        else:
            filters['corpus__in'] = Corpus.objects.readable(self.request.user)

        if 'name' in self.request.query_params:
            filters['name__icontains'] = self.clean_params['name']

        if 'type' in self.request.query_params:
            filters['type__slug'] = self.clean_params['type']

        only_folder = self.clean_params.get('folder')
        if only_folder is not None:
            filters['type__folder'] = only_folder.lower() not in ('false', '0')

        if 'worker_version' in self.clean_params:
            try:
                worker_version_id = uuid.UUID(self.request.query_params['worker_version'])
            except (TypeError, ValueError):
                raise ValidationError({'worker_version': ['Invalid UUID']})

            if not WorkerVersion.objects.filter(id=worker_version_id).exists():
                raise ValidationError({'worker_version': ['This worker version does not exist.']})

            filters['worker_version_id'] = worker_version_id

        return filters

    def get_classifications_filters(self):
        """
        Build Django ORM filters using Q expressions related to best classes.
        Supports 3 modes:
        - elements without any best classes
        - elements with any best classes
        - elements with a specific best class
        """
        class_filter = self.clean_params.get('best_class')
        if class_filter is None:
            return

        # Generic ORM query to find best classes:
        # - elements with a validated classification
        # - OR where high confidence is True
        best_classifications = Q(classifications__state=ClassificationState.Validated) \
            | Q(classifications__high_confidence=True)

        # List elements without any best classes, by inverting the query above
        if class_filter.lower() in ('false', '0'):
            return ~best_classifications

        try:
            # Filter on a specific class
            class_filter = uuid.UUID(class_filter)
            return best_classifications & Q(classifications__ml_class_id=class_filter)
        except (TypeError, ValueError):
            # By default, use all best classifications
            return best_classifications

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['with_zone'] = self.request.query_params.get('with_zone', 'true').lower() not in ('false', '0')
            context['with_corpus'] = self.request.query_params.get('with_corpus', 'true').lower() not in ('false', '0')
        return context

    def get_prefetch(self):
        prefetch = {'zone__image__server'}
        with_best_classes = self.clean_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            prefetch.add(best_classifications_prefetch)

        return prefetch

    def get_order_by(self):
        return ('corpus', 'type__slug', 'name', 'id')

    def filter_queryset(self, queryset):
        # Ordering by ID is required by postgres to order elements with common corpus, type and name
        queryset = queryset \
            .filter(**self.get_filters()) \
            .prefetch_related(*self.get_prefetch()) \
            .select_related('type', 'corpus', 'zone') \
            .order_by(*self.get_order_by())

        class_filters = self.get_classifications_filters()
        if class_filters is not None:
            # Use queryset.distinct() whenever best_class is defined
            queryset = queryset.filter(class_filters).distinct()

        with_has_children = self.clean_params.get('with_has_children')
        if with_has_children and with_has_children.lower() not in ('false', '0'):
            queryset = BulkMap(_fetch_has_children, queryset)

        return queryset

    def check_not_modified(self, queryset):
        """
        Returns True if HTTP 304 Not Modified should be returned. Requires a Django QuerySet instance or a BulkMap.
        """
        # Retrieve the inner QuerySet from the BulkMap if it is used
        if isinstance(queryset, BulkMap):
            queryset = queryset.iterable
        assert isinstance(queryset, QuerySet), 'A Django QuerySet is required to check for modified elements'

        modified_since_string = self.clean_params.get('If-Modified-Since')
        if not modified_since_string:
            return False

        # Parses RFC 2822 dates into tuples
        # Returns None if no date could be parsed or if the header was not defined
        time_tuple = email.utils.parsedate_tz(modified_since_string)
        if not time_tuple:
            raise ValidationError({'If-Modified-Since': ['Bad date format']})

        # Build a datetime from the tuple
        modified_since = datetime.fromtimestamp(email.utils.mktime_tz(time_tuple), timezone.utc)

        # Cause a HTTP 304 only if no element has been created or modified since then
        return not queryset.filter(Q(created__gt=modified_since) | Q(updated__gt=modified_since)).exists()

    def list(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())

        if self.check_not_modified(queryset):
            return Response(None, status=status.HTTP_304_NOT_MODIFIED)

        page = self.paginate_queryset(queryset)
        if page is not None:
            serializer = self.get_serializer(page, many=True)
            return self.get_paginated_response(serializer.data)

        serializer = self.get_serializer(queryset, many=True)
        return Response(serializer.data)

    def delete(self, request, *args, **kwargs):
        queryset = self.filter_queryset(self.get_queryset())
        delete_children = self.clean_params.get('delete_children', '').lower() not in ('false', '0')

        if not queryset.exists():
            raise NotFound

        element_trash(queryset, delete_children=delete_children, user_id=self.request.user.id)
        return Response(status=status.HTTP_204_NO_CONTENT)


class CorpusElements(ElementsListMixin, CorpusACLMixin, DestroyModelMixin, ListAPIView):
    """
    List elements in a corpus and filter by type, name, ML class
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElements',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'top_level',
                'in': 'query',
                'description': 'Only include folder elements without parent elements (top-level elements).',
                'required': False,
                'schema': {'type': 'boolean', 'default': False},
            }
        ]
    })

    @property
    def is_top_level(self):
        return self.clean_params.get('top_level') not in (None, 'false', '0')

    def get_queryset(self):
        # Should not be possible due to the URL
        assert self.selected_corpus, 'Missing corpus ID'
        return self.selected_corpus.elements.all()

    def get_filters(self):
        filters = super().get_filters()

        if self.is_top_level:
            # Only return folders without parents
            filters['paths__isnull'] = True
            filters['type__in'] = self.selected_corpus.types.filter(folder=True)

        return filters

    def get_order_by(self):

        if self.is_top_level:
            return ('name', 'id')

        return super().get_order_by()


class ElementParents(ElementsListMixin, DestroyModelMixin, ACLMixin, ListAPIView):
    """
    List all parents of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementParents',
        'description': 'List all parents of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (parents, grandparents, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    @property
    def is_recursive(self):
        recursive_param = self.clean_params.get('recursive')
        return recursive_param is not None and recursive_param.lower() not in ('false', '0')

    def get_queryset(self):
        role = Role.Guest if self.request.method in permissions.SAFE_METHODS else Role.Contributor

        element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        if not self.has_access(element.corpus, role.value):
            raise PermissionDenied

        return Element.objects.get_ascending(self.kwargs['pk'], recursive=self.is_recursive)


class ElementChildren(ElementsListMixin, DestroyModelMixin, ACLMixin, ListAPIView):
    """
    List all children of an element
    """
    openapi_overrides = ElementsListMixin.openapi_overrides.copy()
    openapi_overrides.update({
        'operationId': 'ListElementChildren',
        'description': 'List all children of an element',
        'parameters': ElementsListMixin.openapi_overrides['parameters'] + [
            {
                'name': 'recursive',
                'in': 'query',
                'description': 'List recursively (children, grandchildren, etc.)',
                'required': False,
                'schema': {
                    'type': 'boolean',
                    'default': False
                }
            },
        ]
    })

    def get_filters(self):
        filters = super().get_filters(corpus=self.element.corpus)
        recursive_param = self.clean_params.get('recursive')

        if recursive_param is None or recursive_param.lower() in ('false', '0'):
            # Only list direct children, by listing all the direct paths from the current element
            # This is faster for direct children than :
            # - Element.objects.filter(paths__path__last=XXX)
            # - Element.objects.get_descending()
            # This code path is extremely used, as it's the basis for navigation amongst elements
            # Be extremely careful when changing this behaviour.
            paths = self.element.paths.values_list('path', flat=True)
            if not paths:
                # Support top level children
                paths = [[]]

            filters['paths__path__in'] = [
                path + [self.kwargs['pk']]
                for path in paths
            ]
        else:
            filters['paths__path__overlap'] = [self.kwargs['pk']]

        return filters

    def get_order_by(self):
        return ('paths__ordering', 'id')

    def get_queryset(self):
        role = Role.Guest if self.request.method in permissions.SAFE_METHODS else Role.Contributor

        self.element = get_object_or_404(
            Element.objects.select_related('corpus'),
            id=self.kwargs['pk']
        )
        if not self.has_access(self.element.corpus, role.value):
            raise PermissionDenied

        return Element.objects.all()


class ElementRetrieve(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Retrieve a single element
    """
    serializer_class = ElementSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_object(self):
        # Prevent duplicating database request
        if not hasattr(self, 'element'):
            self.element = super().get_object()
        return self.element

    def get_queryset(self):
        corpora = Corpus.objects.readable(self.request.user)
        queryset = Element.objects.filter(corpus__in=corpora)
        if self.request and self.request.method == 'DELETE':
            # Only include corpus for ACL check and ID for deletion
            return queryset.select_related('corpus').only('id', 'corpus')

        return queryset \
            .select_related(
                'corpus',
                'type',
                'zone__image__server',
            ) \
            .prefetch_related(Prefetch('classifications', queryset=classifications_queryset)) \
            .annotate(metadata_count=Count('metadatas', filter=Q(metadatas__entity_id=None)))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if self.request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE':
            role = Role.Admin
        if not self.has_access(obj.corpus, role.value):
            access_repr = 'admin' if role == Role.Admin else 'write'
            raise PermissionDenied(detail=f'You do not have {access_repr} access to this element.')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method != 'GET':
            context['element'] = self.get_object()
        return context

    def delete(self, request, *args, **kwargs):
        self.check_object_permissions(self.request, self.get_object())
        queryset = Element.objects.filter(id=self.kwargs['pk'])
        element_trash(queryset, user_id=self.request.user.id)
        return Response(status=status.HTTP_204_NO_CONTENT)


class ElementNeighbors(ListAPIView):
    """
    List neighboring elements
    """
    serializer_class = ElementNeighborsSerializer
    openapi_overrides = {
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'n',
                'in': 'query',
                'description': 'Number of neighbors to retrieve around the element',
                'required': False,
                'schema': {
                    'type': 'integer',
                    'minimum': 1,
                    'maximum': 10,
                },
            },
        ]
    }

    def get_queryset(self):
        n = self.request.query_params.get('n', 1)
        try:
            n = int(n)
        except (TypeError, ValueError):
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})
        if not 1 <= n <= 10:
            raise ValidationError({'n': 'Should be an integer between 1 and 10'})

        try:
            element = Element.objects.get(id=self.kwargs['pk'])
        except Element.DoesNotExist:
            raise NotFound

        return Element.objects.get_neighbors(element, n)


class NoRequestBodySchema(AutoSchema):
    def _get_request_body(self, path, method):
        # Remove request body from OpenAPI schema
        return None


class ElementParent(CreateAPIView, DestroyAPIView):
    """
    Manage element relation with one of its parents
    """
    serializer_class = ElementParentSerializer
    permission_classes = (IsVerified, )
    schema = NoRequestBodySchema()
    openapi_overrides = {
        'tags': ['elements'],
    }

    def get_serializer_from_params(self, child=None, parent=None, **kwargs):
        data = {'child': child, 'parent': parent}
        kwargs['context'] = self.get_serializer_context()
        return ElementParentSerializer(data=data, **kwargs)

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        # Not necessary to serialize again
        serializer.create(serializer.validated_data)
        headers = self.get_success_headers(serializer.data)
        return Response(serializer.data, status=status.HTTP_201_CREATED, headers=headers)

    def destroy(self, request, *args, **kwargs):
        serializer = self.get_serializer_from_params(**kwargs)
        serializer.is_valid(raise_exception=True)
        serializer.delete(serializer.validated_data)
        return Response(status=status.HTTP_204_NO_CONTENT)


class ManageSelection(SelectionMixin, ListAPIView):
    """
    Manage all selected elements
    """
    serializer_class = ElementListSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'ManageSelection',
        'security': [],
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'with_best_classes',
                'in': 'query',
                'description': (
                    'Returns best classifications for each element. '
                    'If not set, elements best_classes field will always be null'
                ),
                'required': False,
                'schema': {
                    'type': 'boolean',
                }
            }
        ]
    }

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        if not settings.ARKINDEX_FEATURES['selection']:
            raise ValidationError(['Selection is not available on this instance.'])

    def get_queryset(self):
        filtered_queryset = self.get_selection().select_related('type', 'corpus', 'zone', 'zone__image', 'zone__image__server').order_by('corpus', 'type__slug', 'name', 'id')

        with_best_classes = self.request.query_params.get('with_best_classes')
        if with_best_classes and with_best_classes.lower() not in ('false', '0'):
            return filtered_queryset.prefetch_related(best_classifications_prefetch)

        return filtered_queryset

    def post(self, request, *args, **kwargs):
        if 'ids' not in request.data:
            raise ValidationError({'ids': ['Element IDs are required.']})

        # Strange field validation done here,
        # as we have different serializers on each request method on the same endpoint.
        # Using serializer fields means we still preserve DRF's normal validation.
        field = serializers.ListField(child=serializers.UUIDField())
        element_ids = field.to_internal_value(request.data['ids'])

        elements = Element.objects.filter(
            id__in=element_ids,
            corpus__in=Corpus.objects.readable(request.user)
        )

        if elements.count() != len(element_ids):
            raise ValidationError({'ids': [
                'Some element IDs do not exist or you do not have permission to access them.'
            ]})

        # Add new Selection objects in bulk, ignoring elements that were already selected
        self.request.user.selected_elements.add(
            *elements.exclude(selection_users=request.user).values_list('id', flat=True)
        )

        return Response(
            status=status.HTTP_201_CREATED,
            data=ElementSlimSerializer(
                elements.prefetch_related('corpus', 'zone__image__server', 'type'),
                context={'request': request},
                many=True
            ).data,
        )

    def delete(self, request, *args, **kwargs):
        if request.data.get('id'):
            field = serializers.PrimaryKeyRelatedField(queryset=self.get_selection())
            element = field.to_internal_value(request.data['id'])
            request.user.selections.get(element=element).delete()
        else:
            request.user.selections.all().delete()

        return Response(status=status.HTTP_204_NO_CONTENT)


class CorpusList(ListCreateAPIView):
    """
    List all available corpora and create new ones
    """
    serializer_class = CorpusSerializer
    pagination_class = None
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        # We need to prefetch the acces rights on each corpora
        accessible_corpora = filter_rights(self.request.user, Corpus, Role.Guest.value)

        # Keep non duplicated corpora with the max access level
        corpora_level = defaultdict(int)
        for c in accessible_corpora:
            if c.max_level > corpora_level[c.id]:
                corpora_level[c.id] = c.max_level

        corpora = Corpus.objects \
            .filter(id__in=corpora_level) \
            .annotate(authorized_users=Count('memberships')) \
            .prefetch_related('types') \
            .order_by('name', 'id')

        for c in corpora:
            c.access_level = corpora_level[c.id]

        return corpora


class CorpusRetrieve(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Get, update and delete corpora
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['corpora'],
    }

    def get_queryset(self):
        return Corpus.objects.readable(self.request.user)

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)

        if self.request.method in permissions.SAFE_METHODS:
            return
        role = Role.Contributor
        if request.method == 'DELETE':
            role = Role.Admin
        if not self.has_access(obj, role.value):
            access_repr = 'admin' if role == Role.Admin else 'write'
            raise PermissionDenied(detail=f'You do not have {access_repr} access to this corpus.')

    def perform_destroy(self, instance):
        corpus_delete(instance, user_id=self.request.user.id)


class TranscriptionsPagination(PageNumberPagination):
    page_size = 100


class ElementTranscriptions(ListAPIView):
    """
    List all transcriptions for an element, optionally filtered by type or worker version id.
    Recursive parameter allow listing transcriptions on sub-elements,
    otherwise element fields in the response will be set to null.
    """
    serializer_class = ElementTranscriptionSerializer
    pagination_class = TranscriptionsPagination
    openapi_overrides = {
        'operationId': 'ListTranscriptions',
        'security': [],
        'tags': ['transcriptions'],
        'parameters': [
            {
                'name': 'recursive',
                'in': 'query',
                'required': False,
                'description': 'Recursively list transcriptions on sub-elements',
                'schema': {'type': 'boolean'}
            },
            {
                'name': 'worker_version',
                'in': 'query',
                'required': False,
                'description': 'Filter transcriptions by worker version',
                'schema': {'type': 'string', 'format': 'uuid'}
            },
            {
                'name': 'element_type',
                'in': 'query',
                'required': False,
                'description': 'Filter transcriptions by the type of their element',
                'schema': {'type': 'string'}
            },
        ]
    }

    @property
    def is_recursive(self):
        if not self.request:
            return
        recursive = self.request.query_params.get('recursive')
        return recursive is not None and recursive.lower() not in ('false', '0')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        # Do serialize the element attached to each transcription in recursive mode only
        context['ignore_element'] = not self.is_recursive
        return context

    def get_queryset(self):
        element = get_object_or_404(Element.objects.filter(
            id=self.kwargs['pk'],
            corpus__in=Corpus.objects.readable(self.request.user)
        ))
        self.check_object_permissions(self.request, element)

        # ORDER BY casting IDs as char to avoid the PostgreSQL optimizer's inefficient scan
        # TODO: See if select_related is faster than a prefetch on this endpoint
        queryset = Transcription.objects \
            .prefetch_related('element__zone__image__server', 'worker_version') \
            .annotate(char_id=Cast('id', output_field=CharField())) \
            .order_by('char_id')

        if self.is_recursive:
            queryset = queryset.filter(
                # Retrieve both element and sub-elements transcriptions
                element__in=[element.id, *Element.objects.get_descending(element.id).values_list('id')]
            ).prefetch_related('element__type', 'element__zone')
        else:
            queryset = queryset.filter(element_id=element.id)

        return queryset

    def filter_queryset(self, queryset):
        # Filter by worker version
        worker_version = self.request.query_params.get('worker_version')
        if worker_version:
            queryset = queryset.filter(worker_version_id=worker_version)

        # Filter by element_type
        element_type = self.request.query_params.get('element_type')
        if element_type:
            queryset = queryset.select_related('element__type').filter(element__type__slug=element_type)

        return queryset


class ElementsCreate(CreateAPIView):
    """
    Create a new element
    """
    permission_classes = (IsVerified, )
    serializer_class = ElementCreateSerializer
    openapi_overrides = {
        'operationId': 'CreateElement',
        'tags': ['elements'],
        'parameters': [
            {
                'name': 'slim_output',
                'in': 'query',
                'required': False,
                'description': 'When enabled, the endpoint will only return the new Element\'s ID. This is especially useful for an automated tool, as it speeds up the element creation.',
                'schema': {
                    'type': 'boolean',
                }
            },
        ]
    }


class ElementMetadata(ListCreateAPIView):
    """
    List all metadata linked to an element.


    Create a metadata on an existing element.

    The `date` type allows dates to be parsed and indexed for search APIs. The supported date formats are as follows:

    * `YYYY`: `1995` (recommended)
    * `YYYY-MM`: `1995-02` (recommended)
    * `YYYY-MM-DD`: `1995-02-16` (recommended)
    * `/YYYY(.*)month/`: `1995, february`, with months in French or English (case-insensitive)
    * `YYYY-YYYY`: `1995-1996` (yields a date interval)

    The `reference` type is also indexed in search APIs and allows searching by
    known identifiers or references, or any arbitrary string.
    """
    permission_classes = (IsVerified, )
    pagination_class = None
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'tags': ['elements']
    }

    def get_queryset(self):
        if self.request and self.request.method == 'GET':
            element = get_object_or_404(Element, id=self.kwargs['pk'], corpus__in=Corpus.objects.readable(self.request.user))
            return element.metadatas.filter(entity_id=None).select_related('revision__repo')

        # We need to use the default database to avoid stale read
        # when a metadata is created immediately after an element creation
        return Element.objects.using('default').filter(corpus__in=Corpus.objects.writable(self.request.user))

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request and self.request.method == 'GET':
            context['list'] = True
        elif self.request and self.request.method != 'GET':  # Ignore this step when generating the schema with OpenAPI
            context['element'] = self.get_object()
        return context


class MetadataEdit(ACLMixin, RetrieveUpdateDestroyAPIView):
    """
    Edit an existing element medadata
    """
    permission_classes = (IsVerified, )
    serializer_class = MetaDataUpdateSerializer
    openapi_overrides = {
        'tags': ['elements']
    }

    def get_queryset(self):
        # Filter readable metadata in order to check object permissions
        return MetaData.objects.select_related('element__corpus').filter(element__corpus__in=Corpus.objects.readable(self.request.user))

    def check_object_permissions(self, request, obj):
        super().check_object_permissions(request, obj)
        if not self.has_access(obj.element.corpus, Role.Contributor.value):
            self.permission_denied(request, message='You do not have write access to this corpus.')

    def perform_destroy(self, instance):
        serializer = self.get_serializer(instance)
        serializer.delete(instance)


class AllowedMetaDataPagination(PageNumberPagination):
    page_size = 200


class CorpusAllowedMetaData(CorpusACLMixin, ListAPIView):
    """
    List allowed metadata (type, name) couples in a corpus
    """
    serializer_class = CorpusAllowedMetaDataSerializer
    pagination_class = AllowedMetaDataPagination
    openapi_overrides = {
        'operationId': 'ListCorpusAllowedMetaDatas',
        'description': 'List (type, name) couples allowed for corpus metadata',
        'tags': ['corpora']
    }

    def get_queryset(self):
        if not self.request:
            return AllowedMetaData.objects.none()
        return AllowedMetaData.objects.filter(corpus=self.get_corpus(self.kwargs['pk']))


class ElementTypeCreate(CreateAPIView):
    """
    Create a new element type for a specific corpus
    """
    serializer_class = ElementTypeSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'tags': ['elements']
    }


class ElementTypeUpdate(UpdateAPIView):
    """
    Update an existing element type
    """
    serializer_class = ElementTypeLightSerializer
    permission_classes = (IsVerifiedOrReadOnly, )
    openapi_overrides = {
        'operationId': 'UpdateElementType',
        'tags': ['elements']
    }

    def get_queryset(self):
        return ElementType.objects.filter(corpus__in=Corpus.objects.readable(self.request.user))


class ElementBulkCreate(CreateAPIView):
    """
    Create multiple child elements at once on a single parent
    """
    serializer_class = ElementBulkSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'CreateElements',
        'tags': ['elements'],
        'responses': {
            '201': {
                'description': 'Created',
                'content': {
                    'application/json': {
                        'examples': [{
                            'value': [{
                                'id': 'string'
                            }]
                        }]
                    }
                }
            }
        }
    }

    def get_object(self):
        if not hasattr(self, 'element'):
            self.element = super().get_object()
            if not self.element.zone_id:
                raise ValidationError({'element': ['Cannot create child elements on an element without a zone.']})
        return self.element

    def get_queryset(self):
        if not self.request:
            return Element.objects.none()
        return Element \
            .objects \
            .filter(corpus__in=Corpus.objects.writable(self.request.user)) \
            .select_related('zone') \
            .only('id', 'corpus_id', 'zone__id', 'zone__image_id')

    def get_serializer_context(self):
        context = super().get_serializer_context()
        if self.request:
            context['element'] = self.get_object()
        return context

    def create(self, request, *args, **kwargs):
        serializer = self.get_serializer(data=request.data)
        serializer.is_valid(raise_exception=True)
        data = self.perform_create(serializer)
        headers = self.get_success_headers({})
        return Response(data, status=status.HTTP_201_CREATED, headers=headers)

    @transaction.atomic
    def perform_create(self, serializer):
        worker_version = serializer.validated_data.get('worker_version')
        elements = serializer.validated_data['elements']

        # Retrieve existing zones from the DB
        image_id = self.element.zone.image_id
        existing_zones = {
            # Use WKB representation to compare existing zones
            # to avoid comparing references or slower coordinates
            polygon.wkb: zone_id
            for polygon, zone_id in Zone.objects.using('default').filter(image_id=image_id).values_list('polygon', 'id')
        }

        # Retrieve or create required zones
        new_zones = []
        for element_data in elements:
            polygon = element_data.pop('polygon')
            # Retrieve or create the zone on the image
            zone_id = existing_zones.get(polygon.wkb)
            if not zone_id:
                new_zone = Zone(
                    id=uuid.uuid4(),
                    image_id=image_id,
                    polygon=polygon,
                )
                zone_id = new_zone.id
                existing_zones[polygon.wkb] = zone_id
                new_zones.append(new_zone)
            element_data['zone'] = zone_id

        Zone.objects.bulk_create(new_zones)

        # ElementPath.get_next_order, but for multiple types
        next_orderings = defaultdict(
            int,
            ElementPath
            .objects
            .using('default')
            .filter(
                path__last=self.element.id,
                element__type_id__in=set(element_data['type'] for element_data in elements)
            )
            .values('element__type_id')
            .annotate(next_ordering=Max('ordering') + 1)
            .values_list('element__type_id', 'next_ordering')
        )
        # Prepare elements
        for element_data in elements:
            # Keep ordering in the element payload to optimize element path creation later
            element_data['ordering'] = next_orderings[element_data['type']]
            next_orderings[element_data['type']] += 1

            element_data['element'] = Element(
                id=uuid.uuid4(),
                zone_id=element_data['zone'],
                corpus_id=self.element.corpus_id,
                type_id=element_data['type'],
                name=element_data['name'],
                worker_version=worker_version,
            )

        Element.objects.bulk_create([
            element_data['element'] for element_data in elements
        ])

        # Support top level elements, by adding an empty initial path to trigger loops below
        parent_paths = list(self.element.paths.all().values_list('path', flat=True))
        if not parent_paths:
            parent_paths = [[]]

        ElementPath.objects.bulk_create([
            ElementPath(
                path=parent_path + [self.element.id],
                element_id=element_data['element'].id,
                ordering=element_data['ordering'],
            )
            for parent_path in parent_paths
            for element_data in elements
        ])

        return [{'id': element_data['element'].id} for element_data in elements]


class CorpusDeleteSelection(CorpusACLMixin, SelectionMixin, DestroyAPIView):
    """
    Delete all selected elements on a corpus
    """
    serializer_class = CorpusSerializer
    permission_classes = (IsVerified, )
    openapi_overrides = {
        'operationId': 'DeleteCorpusSelection',
        'tags': ['elements']
    }

    def delete(self, request, *args, **kwargs):
        corpus = self.get_corpus(self.kwargs['pk'], role=Role.Admin)
        selected_elements = self.get_selection(corpus_id=corpus.id)
        if not selected_elements.exists():
            raise NotFound

        for batch in range(0, selected_elements.count(), 50):
            queryset = Element.objects.filter(id__in=list(selected_elements[batch:batch + 50].values_list('id', flat=True)))
            element_trash(queryset, user_id=self.request.user.id)

        return Response(status=status.HTTP_204_NO_CONTENT)
