from django.db.models import Max
from rest_framework import serializers
from rest_framework.exceptions import APIException, ValidationError

from arkindex.dataimport.serializers.git import RevisionSerializer
from arkindex.documents.dates import DateType
from arkindex.documents.models import AllowedMetaData, Corpus, Element, ElementType, MetaData
from arkindex.images.serializers import ZoneLightSerializer
from arkindex.project.serializer_fields import EnumField
from arkindex.project.triggers import reindex_start
from arkindex_common.enums import MetaType


class InterpretedDateSerializer(serializers.Serializer):
    """
    Serialize a list of interpreted dates linked to a metadata
    """
    type = EnumField(DateType)
    year = serializers.IntegerField(min_value=0)
    month = serializers.IntegerField(min_value=1, max_value=12, allow_null=True)
    day = serializers.IntegerField(min_value=1, max_value=31, allow_null=True)

    class Meta:
        fields = (
            'year',
            'month',
            'day',
            'type',
        )


class ElementLightSerializer(serializers.ModelSerializer):
    """
    Serialises a Element
    """
    type = serializers.CharField(source='type.slug')

    class Meta:
        model = Element
        fields = (
            'id',
            'type',
            'name',
        )


class ElementZoneSerializer(ElementLightSerializer):
    """
    Lightly serialises an element with its type and zone
    """
    zone = ZoneLightSerializer()

    class Meta(ElementLightSerializer.Meta):
        model = Element
        fields = ElementLightSerializer.Meta.fields + ('zone', )


class ElementTypeLightSerializer(serializers.ModelSerializer):

    class Meta:
        model = ElementType
        fields = (
            'id',
            'slug',
            'display_name',
            'folder',
        )

    def validate(self, data):
        slug = data.get('slug') if 'slug' in data else self.instance.slug
        corpus = data.get('corpus') if 'corpus' in data else self.instance.corpus
        type = corpus.types.filter(slug=slug)
        if type.exists() and (not self.instance or type.first().id != self.instance.id):
            raise ValidationError({'slug': ['Slug must be unique']})
        return data


class CorpusLightSerializer(serializers.ModelSerializer):
    """
    Serializes a Corpus
    """

    class Meta:
        model = Corpus
        fields = (
            'id',
            'name',
            'public',
        )


class DuplicatedMetadata(APIException):
    status_code = 200
    default_detail = 'Metadata exists.'


class MetaDataLightSerializer(serializers.ModelSerializer):
    """
    Serialize a metadata without its entity
    """
    type = EnumField(MetaType)
    revision = RevisionSerializer(read_only=True)
    dates = InterpretedDateSerializer(many=True, source='get_dates', read_only=True)

    def reindex_element(self, elt):
        reindex_start(element=elt, elements=True)

    def check_allowed(self, corpus=None, type=None, name=None, instance=None, **kwargs):
        """
        Assert metadata type and value are allowed in this corpus
        """
        message = 'Metadata with this (type, name) are forbidden for this corpus'
        if instance:
            type = instance.type
            name = instance.name
            message = 'Edition of metadata with this (type, name) are disallowed for this corpus'
        if not AllowedMetaData.objects.filter(corpus=corpus, type=type, name=name):
            raise ValidationError({
                '__all__': [message]
            })

    def create(self, validated_data):
        user = self.context['request'].user
        element = self.context['element']
        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=element.corpus, **validated_data)
        metadata = element.metadatas.using('default').filter(name=validated_data['name'])
        if metadata.exists():
            if metadata.filter(value=validated_data['value'], type=validated_data['type']).exists():
                # Raise a 200_OK if a metadata with same type, name and value exists
                raise DuplicatedMetadata
            validated_data['index'] = metadata.aggregate(Max('index'))['index__max'] + 1

        metadata = element.metadatas.create(**validated_data)
        if metadata.type == MetaType.Reference:
            # Index reference metadatas
            self.reindex_element(element)
        return metadata

    def update(self, instance, validated_data):
        user = self.context['request'].user
        element = instance.element
        if not (user.is_admin or user.is_internal):
            # Assert actual instance is part of AllowedMetaData
            self.check_allowed(corpus=element.corpus, instance=instance)
            # Assert update is part of AllowedMetaData
            self.check_allowed(corpus=element.corpus, **validated_data)
        new_name = validated_data.get('name') or instance.name
        metadata = element.metadatas \
            .filter(name=new_name) \
            .exclude(id=instance.id)
        if metadata.exists():
            validated_data['index'] = metadata.aggregate(Max('index'))['index__max'] + 1

        original_type = instance.type
        updated_metadata = super().update(instance, validated_data)
        if MetaType.Reference in (original_type, updated_metadata.type):
            # Reindex element reference metadatas
            self.reindex_element(updated_metadata.element)
        return updated_metadata

    def delete(self, instance):
        user = self.context['request'].user
        element, meta_type = instance.element, instance.type
        if not (user.is_admin or user.is_internal):
            self.check_allowed(corpus=element.corpus, instance=instance)

        instance.delete()
        if meta_type == MetaType.Reference:
            # Reindex metadata's element
            self.reindex_element(element)

    class Meta:
        model = MetaData
        read_only_fields = ('id', )
        fields = (
            'id',
            'type',
            'name',
            'value',
            'revision',
            'dates',
        )


class CorpusAllowedMetaDataSerializer(serializers.ModelSerializer):
    """
    Serialize allowed (type, name) couples for metadata creation in a corpus
    """
    type = EnumField(MetaType)

    class Meta:
        model = AllowedMetaData
        fields = ('type', 'name', )
        read_only_fields = ('type', 'name', )
