import math

from django.conf import settings
from rest_framework import serializers

from arkindex.documents.date_parser import parse_date
from arkindex.documents.models import Element, ElementType, Entity
from arkindex.documents.serializers.light import CorpusLightSerializer, ElementLightSerializer
from arkindex.documents.serializers.ml import TranscriptionSerializer
from arkindex.images.serializers import ZoneSerializer
from arkindex.project.serializer_fields import EnumField
from arkindex_common.enums import EntityType


class SearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query

    date_lte is rounded to superior value using ES syntax (see round_date()) and served as date_lt (exactly lower than)
    Rounded value depend if month/day are provided (Ex: '<= march 1999' becomes '< april 1999')
    """
    q = serializers.CharField(source='query', default=None, max_length=settings.ES_QUERY_STRING_MAX_LENGTH)
    score = serializers.FloatField(source='min_score', min_value=0.0, max_value=1.0, default=0.0)
    date_gte = serializers.CharField(default=None)
    date_lte = serializers.CharField(source='date_lt', default=None)
    element_type = serializers.SlugField(default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)

    def parse_date(self, raw_date):
        date = parse_date(raw_date)
        if len(date) != 1:
            raise serializers.ValidationError('Could not parse Date')
        return date[0]

    def validate_score(self, value):
        if not isinstance(value, float) or math.isinf(value) or math.isnan(value):
            raise serializers.ValidationError('Score should be a valid decimal number between 0 and 1')
        return value

    def validate_date_gte(self, value):
        if not value:
            return
        return self.parse_date(value).es_str()

    def validate_date_lte(self, value):
        if not value:
            return
        return self.parse_date(value).es_round_up()

    def validate(self, data):
        data = super().validate(data)

        gte, lt = data['date_gte'], data['date_lt']
        if gte and lt and gte > lt:
            raise serializers.ValidationError("Upper date must be greater than lower date")
        if not (data['query'] or data['date_lt'] or data['date_gte']):
            raise serializers.ValidationError("At least query terms or dates must be provided")

        corpus_id, element_type = data['corpus_id'], data['element_type']
        if element_type:
            if not corpus_id:
                raise serializers.ValidationError('A corpus is required to filter by element types')
            if not ElementType.objects.filter(corpus_id=corpus_id, slug=element_type).exists():
                raise serializers.ValidationError(
                    'Type with slug {!r} does not exist in corpus {}'.format(element_type, corpus_id),
                )

        return data


class EntitySearchQuerySerializer(serializers.Serializer):
    """
    Search parameters validation serializer in order to build a ES query for entity
    """
    q = serializers.CharField(source='query', max_length=settings.ES_QUERY_STRING_MAX_LENGTH)
    type = EnumField(enum=EntityType, default=None)
    corpus = serializers.UUIDField(source='corpus_id', default=None)


class IIIFSearchQuerySerializer(serializers.Serializer):
    """
    Search parameters for IIIF transcription search
    See https://iiif.io/api/search/1.0/#request-1
    """
    q = serializers.CharField(source='query', max_length=settings.ES_QUERY_STRING_MAX_LENGTH)


class ElementSearchResultSerializer(ElementLightSerializer):
    """
    A page search result with nested transcriptions
    """
    zone = ZoneSerializer()
    transcriptions = TranscriptionSerializer(many=True, source='transcriptions_results')
    total_transcriptions = serializers.IntegerField()
    parent_paths = serializers.ListField(
        child=serializers.ListField(
            child=ElementLightSerializer()
        ),
    )
    corpus = CorpusLightSerializer()
    date_from = serializers.CharField()
    date_to = serializers.CharField()

    class Meta:
        model = Element
        fields = (
            'id',
            'name',
            'type',
            'zone',
            'transcriptions',
            'total_transcriptions',
            'parent_paths',
            'corpus',
            'date_from',
            'date_to',
        )


class EntitySearchResultSerializer(serializers.ModelSerializer):
    """
    Serializes an entity
    """
    type = EnumField(EntityType)

    class Meta:
        model = Entity
        fields = (
            'id',
            'name',
            'type',
            'metas'
        )
