from django.conf import settings
from django.core.exceptions import PermissionDenied
from django.db.models import Q
from django.shortcuts import get_object_or_404
from django.views.decorators.cache import cache_page
from rest_framework.exceptions import APIException, ValidationError
from rest_framework.serializers import Serializer

from arkindex.dataimport.models import Repository
from arkindex.documents.models import Corpus
from arkindex.documents.serializers.search import SearchQuerySerializer
from arkindex.project.elastic import ESQuerySet
from arkindex.project.openapi import AutoSchema, SearchAutoSchema
from arkindex.project.pagination import CustomCursorPagination
from arkindex.users.models import Role
from arkindex.users.utils import check_level_param, filter_rights


class ACLMixin(object):
    """
    Access control mixin using the generic Right table.
    """
    _user = None

    def __init__(self, user=None):
        self._user = user

    @property
    def user(self):
        return self._user or self.request.user

    def has_access(self, instance, level, skip_public=False):
        """
        Check if the user has access to a generic instance with a minimum level
        If skip_public parameter is set to true, exclude rights on public instances
        """
        check_level_param(level)

        # Handle special authentications
        if not skip_public and level <= Role.Guest.value and getattr(instance, 'public', False):
            return True
        if self.user.is_anonymous:
            return False
        elif self.user.is_admin or self.user.is_internal:
            return True

        return instance.memberships.filter(
            Q(
                # Right direcly owned by this user
                Q(user=self.user)
                & Q(level__gte=level)
            )
            | Q(
                # Right owned by the group and by the user
                Q(group__memberships__user=self.user)
                & Q(level__gte=level)
                & Q(group__memberships__level__gte=level)
            )
        ).exists()


class RepositoryACLMixin(ACLMixin):

    @property
    def readable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Guest.value).values('id')
        )

    @property
    def executable_repositories(self):
        return Repository.objects.filter(
            id__in=filter_rights(self.user, Repository, Role.Contributor.value).values('id')
        )

    def has_read_access(self, repo):
        return self.has_access(repo, Role.Guest.value)

    def has_execution_access(self, repo):
        return self.has_access(repo, Role.Contributor.value)

    def has_admin_access(self, repo):
        return self.has_access(repo, Role.Admin.value)


class CorpusACLMixin(ACLMixin):

    def get_corpus(self, corpus_id, role=Role.Guest):
        """
        Try to retrieve a corpus according to user rights.
        Raises a 404 if the corpus does not exists and a 403 if the permissions are invalid
        """
        corpus = get_object_or_404(Corpus, id=corpus_id)
        if not self.has_access(corpus, role.value):
            raise PermissionDenied
        return corpus

    @property
    def readable_corpora(self):
        return Corpus.objects.filter(
            id__in=filter_rights(self.user, Corpus, Role.Guest.value).values('id')
        )

    @property
    def writable_corpora(self):
        return Corpus.objects.filter(
            id__in=filter_rights(self.user, Corpus, Role.Contributor.value).values('id')
        )

    def has_read_access(self, corpus):
        return self.has_access(corpus, Role.Guest.value)

    def has_write_access(self, corpus):
        return self.has_access(corpus, Role.Contributor.value)

    def has_admin_access(self, corpus):
        return self.has_access(corpus, Role.Admin.value)


class SearchAPIMixin(CorpusACLMixin):
    query_serializer_class = SearchQuerySerializer
    schema = SearchAutoSchema()
    search = None

    def get_search(self, **query):
        return self.search

    def get_queryset(self):
        if not settings.ARKINDEX_FEATURES['search']:
            raise ValidationError(['Search features are not available on this instance.'])

        serializer = self.query_serializer_class(
            context={'request': self.request},
            data=self.request.query_params,
        )
        serializer.is_valid(raise_exception=True)
        query = serializer.validated_data
        # TODO Handle corpus field in serializer too
        if query.get('corpus_id'):
            try:
                query['corpora_ids'] = [str(self.get_corpus(query['corpus_id']).id), ]
            except Corpus.DoesNotExist:
                raise PermissionDenied
        else:
            query['corpora_ids'] = list(map(
                str,
                Corpus.objects.readable(self.request.user).values_list('id', flat=True),
            ))

        if not query['corpora_ids']:
            return []

        query.pop('corpus_id', None)
        search = self.get_search(**query)
        return ESQuerySet(self.post_process, search)

    def post_process(self, *args):
        return args


class SelectionMixin(object):

    def get_selection(self, corpus_id=None):
        assert settings.ARKINDEX_FEATURES['selection'], 'Selection feature is unavailable'
        queryset = self.request.user.selected_elements.all()
        if corpus_id is not None:
            queryset = queryset.filter(corpus_id=corpus_id)
        return queryset


class DeprecatedAPIException(APIException):
    status_code = 410
    default_detail = 'This endpoint has been deprecated.'
    default_code = 'deprecated'


class DeprecatedAutoSchema(AutoSchema):
    """
    OpenAPI schema for deprecated views.
    """
    def _get_responses(self, path, method):
        """
        Return nothing but the deprecation error response.
        """
        deprecation_message = getattr(self.view, 'deprecation_message', None) or DeprecatedAPIException.default_detail

        schema = {
            'schema': {
                'type': 'object',
                'properties': {
                    'detail': {
                        'type': 'string',
                        'const': deprecation_message
                    }
                },
                'example': {
                    'detail': deprecation_message
                }
            }
        }

        return {
            str(DeprecatedAPIException.status_code): {
                'content': {
                    media_type: schema
                    for media_type in self.map_renderers(path, method)
                },
                'description': 'Deprecation message'
            }
        }

    def get_operation(self, path, method):
        operation = super().get_operation(path, method)
        operation['deprecated'] = True
        return operation


class DeprecatedMixin(object):
    """
    Add this mixin to an APIView to make it deprecated.
    """
    schema = DeprecatedAutoSchema()
    # Add a default useless serializer because OpenAPI generation requires all endpoints to have one
    serializer_class = Serializer
    deprecation_message = None

    def initial(self, request, *args, **kwargs):
        super().initial(request, *args, **kwargs)
        method = request.method.lower()
        if method in self.http_method_names and hasattr(self, method):
            raise DeprecatedAPIException(detail=getattr(self, 'deprecation_message', None))


class CachedViewMixin(object):
    """
    Add this mixin to any class-based view to cache it.
    """
    cache_timeout = None
    cache_prefix = None

    def __init__(self, *args, **kwargs):
        if self.cache_timeout:
            self.dispatch = cache_page(
                self.cache_timeout,
                key_prefix=self.cache_prefix,
            )(self.dispatch)


class CustomPaginationViewMixin(object):
    """
    A custom cursor pagination mixin
    Elements count can be retrieved with the `with_count` parameter
    """
    pagination_class = CustomCursorPagination

    @property
    def paginator(self):
        if not hasattr(self, '_paginator'):
            params = self.request.query_params
            with_count = params.get('with_count') not in (None, '', 'false', '0')
            self._paginator = self.pagination_class(with_count=with_count)
        return self._paginator
